// Copyright (c) The Geekeey Authors
// SPDX-License-Identifier: EUPL-1.2

namespace Geekeey.Extensions.Result.Tests;

[TestFixture]
internal sealed class ResultTransformTests
{
	[Test]
	public void I_can_map_and_it_returns_success_for_success()
	{
		var start = Success(2);
		var result = start.Map(value => value.ToString());

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.True);
			Assert.That(result.Value, Is.EqualTo("2"));
		});
	}

	[Test]
	public void I_can_map_and_it_returns_failure_for_failure()
	{
		var start = Failure<int>("error");
		var result = start.Map(value => value.ToString());

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public void I_can_transform_result_with_then_and_it_returns_success_for_success_and_mapping_returning_success()
	{
		var start = Success(2);
		var result = start.Then(value => Success(value.ToString()));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.True);
			Assert.That(result.Value, Is.EqualTo("2"));
		});
	}

	[Test]
	public void I_can_transform_result_with_then_and_it_returns_failure_for_success_and_mapping_returning_failure()
	{
		var start = Success(2);
		var result = start.Then(_ => Failure<string>("error"));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public void I_can_transform_result_with_then_and_it_returns_failure_for_failure_and_mapping_returning_success()
	{
		var start = Failure<int>("error");
		var result = start.Then(value => Success(value.ToString()));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public void I_can_transform_result_with_then_and_it_returns_failure_for_failure_and_mapping_returning_failure()
	{
		var start = Failure<int>("error");
		var result = start.Then(_ => Failure<int>("error 2"));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public void I_can_try_map_and_it_returns_success_for_success_without_throwing()
	{
		var start = Success(2);
		var result = start.TryMap(value => value.ToString());

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.True);
			Assert.That(result.Value, Is.EqualTo("2"));
		});
	}

	[Test]
	public void I_can_try_map_and_it_returns_failure_for_failure_without_throwing()
	{
		var start = Failure<int>("error");
		var result = start.TryMap(value => value.ToString());

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public void I_can_try_map_and_it_returns_failure_for_success_with_throwing()
	{
		var start = Success(2);
		var result = start.TryMap<string>(_ => throw new CustomTestException());

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<ExceptionError>());
			var instance = result.Error as ExceptionError;
			Assert.That(instance?.Exception, Is.InstanceOf<CustomTestException>());
		});
	}

	[Test]
	public void I_can_try_map_and_it_returns_failure_for_failure_with_throwing()
	{
		var start = Failure<int>("error");
		var result = start.TryMap<string>(_ => throw new CustomTestException());

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<StringError>());
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public void I_can_try_transform_result_with_then_and_it_returns_success_for_success_and_mapping_returning_success()
	{
		var start = Success(2);
		var result = start.ThenTry(value => Success(value.ToString()));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.True);
			Assert.That(result.Value, Is.EqualTo("2"));
		});
	}

	[Test]
	public void I_can_try_transform_result_with_then_and_it_returns_failure_for_success_and_mapping_returning_failure()
	{
		var start = Success(2);
		var result = start.ThenTry(_ => Failure<string>("error"));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public void I_can_try_transform_result_with_then_and_it_returns_failure_for_failure_and_mapping_returning_failure()
	{
		var start = Failure<int>("error");
		var result = start.ThenTry(x => Success(x.ToString()));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public void I_can_try_transform_result_with_then_and_it_returns_failure_for_success_and_mapping_throwing()
	{
		var start = Success(2);
		var result = start.ThenTry<string>(_ => throw new CustomTestException());

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<ExceptionError>());
			var instance = result.Error as ExceptionError;
			Assert.That(instance?.Exception, Is.InstanceOf<CustomTestException>());
		});
	}

	[Test]
	public void I_can_try_transform_result_with_then_and_it_returns_failure_for_failure_and_mapping_throwing()
	{
		var start = Failure<int>("error");
		var result = start.ThenTry<string>(_ => throw new CustomTestException());

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<StringError>());
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task I_can_map_async_and_it_returns_success_for_success()
	{
		var start = Success(2);
		var result = await start.MapAsync(value => Task.FromResult(value.ToString()));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.True);
			Assert.That(result.Value, Is.EqualTo("2"));
		});
	}

	[Test]
	public async Task I_can_map_async_and_it_returns_failure_for_failure()
	{
		var start = Failure<int>("error");
		var result = await start.MapAsync(value => Task.FromResult(value.ToString()));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task I_can_transform_result_async_with_then_and_it_returns_success_for_success_and_mapping_returning_success()
	{
		var start = Success(2);
		var result = await start.ThenAsync(value => Task.FromResult(Success(value.ToString())));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.True);
			Assert.That(result.Value, Is.EqualTo("2"));
		});
	}

	[Test]
	public async Task I_can_transform_result_async_with_then_and_it_returns_failure_for_success_and_mapping_returning_failure()
	{
		var start = Success(2);
		var result = await start.ThenAsync(_ => Task.FromResult(Failure<string>("error")));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task I_can_transform_result_async_with_then_and_it_returns_failure_for_failure_and_mapping_returning_success()
	{
		var start = Failure<int>("error");
		var result = await start.ThenAsync(value => Task.FromResult(Success(value.ToString())));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task I_can_transform_result_async_with_then_and_it_returns_failure_for_failure_and_mapping_returning_failure()
	{
		var start = Failure<int>("error");
		var result = await start.ThenAsync(_ => Task.FromResult(Failure<int>("error 2")));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task I_can_try_map_async_and_it_returns_success_for_success_without_throwing()
	{
		var start = Success(2);
		var result = await start.TryMapAsync(value => Task.FromResult(value.ToString()));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.True);
			Assert.That(result.Value, Is.EqualTo("2"));
		});
	}

	[Test]
	public async Task I_can_try_map_async_and_it_returns_failure_for_failure_without_throwing()
	{
		var start = Failure<int>("error");
		var result = await start.TryMapAsync(value => Task.FromResult(value.ToString()));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task I_can_try_map_async_and_it_returns_failure_for_success_with_throwing()
	{
		var start = Success(2);
		var result = await start.TryMapAsync(Task<string> (_) => throw new CustomTestException());

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<ExceptionError>());
			var instance = result.Error as ExceptionError;
			Assert.That(instance?.Exception, Is.InstanceOf<CustomTestException>());
		});
	}

	[Test]
	public async Task I_can_try_map_async_and_it_returns_failure_for_success_with_await_throwing()
	{
		var start = Success(2);
		var result = await start.TryMapAsync(async Task<string> (_) =>
		{
			await Task.CompletedTask;
			throw new CustomTestException();
		});

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<ExceptionError>());
			var instance = result.Error as ExceptionError;
			Assert.That(instance?.Exception, Is.InstanceOf<CustomTestException>());
		});
	}

	[Test]
	public async Task I_can_try_map_async_and_it_returns_failure_for_failure_with_throwing()
	{
		var start = Failure<int>("error");
		var result = await start.TryMapAsync(Task<string> (_) => throw new CustomTestException());

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<StringError>());
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task I_can_try_map_async_and_it_returns_failure_for_failure_with_await_throwing()
	{
		var start = Failure<int>("error");
		var result = await start.TryMapAsync(async Task<string> (_) =>
		{
			await Task.CompletedTask;
			throw new CustomTestException();
		});

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<StringError>());
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task I_can_try_transform_result_async_with_then_and_it_returns_success_for_success_and_mapping_returning_success()
	{
		var start = Success(2);
		var result = await start.ThenTryAsync(value => Task.FromResult(Success(value.ToString())));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.True);
			Assert.That(result.Value, Is.EqualTo("2"));
		});
	}

	[Test]
	public async Task I_can_try_transform_result_async_with_then_and_it_returns_failure_for_success_and_mapping_returning_failure()
	{
		var start = Success(2);
		var result = await start.ThenTryAsync(_ => Task.FromResult(Failure<string>("error")));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task I_can_try_transform_result_async_with_then_and_it_returns_failure_for_failure_and_mapping_returning_failure()
	{
		var start = Failure<int>("error");
		var result = await start.ThenTryAsync(x => Task.FromResult(Success(x.ToString())));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task I_can_try_transform_result_async_with_then_and_it_returns_failure_for_success_and_mapping_throwing()
	{
		var start = Success(2);
		var result = await start.ThenTryAsync(Task<Result<string>> (_) => throw new CustomTestException());

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<ExceptionError>());
			var instance = result.Error as ExceptionError;
			Assert.That(instance?.Exception, Is.InstanceOf<CustomTestException>());
		});
	}

	[Test]
	public async Task I_can_try_transform_result_async_with_then_and_it_returns_failure_for_success_and_mapping_await_throwing()
	{
		var start = Success(2);
		var result = await start.ThenTryAsync(async Task<Result<string>> (_) =>
		{
			await Task.CompletedTask;
			throw new CustomTestException();
		});

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<ExceptionError>());
			var instance = result.Error as ExceptionError;
			Assert.That(instance?.Exception, Is.InstanceOf<CustomTestException>());
		});
	}

	[Test]
	public async Task I_can_try_transform_result_async_with_then_and_it_returns_failure_for_failure_and_mapping_throwing()
	{
		var start = Failure<int>("error");
		var result = await start.ThenTryAsync(Task<Result<string>> (_) => throw new CustomTestException());

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<StringError>());
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task I_can_try_transform_result_async_with_then_and_it_returns_failure_for_failure_and_mapping_await_throwing()
	{
		var start = Failure<int>("error");
		var result = await start.ThenTryAsync(async Task<Result<string>> (_) =>
		{
			await Task.CompletedTask;
			throw new CustomTestException();
		});

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<StringError>());
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task I_can_map_async_and_it_returns_success_for_success_ValueTask()
	{
		var start = Success(2);
		var result = await start.MapAsync(value => ValueTask.FromResult(value.ToString()));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.True);
			Assert.That(result.Value, Is.EqualTo("2"));
		});
	}

	[Test]
	public async Task I_can_map_async_and_it_returns_failure_for_failure_ValueTask()
	{
		var start = Failure<int>("error");
		var result = await start.MapAsync(value => ValueTask.FromResult(value.ToString()));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task I_can_transform_result_async_with_then_and_it_returns_success_for_success_and_mapping_returning_success_ValueTask()
	{
		var start = Success(2);
		var result = await start.ThenAsync(value => ValueTask.FromResult(Success(value.ToString())));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.True);
			Assert.That(result.Value, Is.EqualTo("2"));
		});
	}

	[Test]
	public async Task I_can_transform_result_async_with_then_and_it_returns_failure_for_success_and_mapping_returning_failure_ValueTask()
	{
		var start = Success(2);
		var result = await start.ThenAsync(_ => ValueTask.FromResult(Failure<string>("error")));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task I_can_transform_result_async_with_then_and_it_returns_failure_for_failure_and_mapping_returning_success_ValueTask()
	{
		var start = Failure<int>("error");
		var result = await start.ThenAsync(value => ValueTask.FromResult(Success(value.ToString())));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task I_can_transform_result_async_with_then_and_it_returns_failure_for_failure_and_mapping_returning_failure_ValueTask()
	{
		var start = Failure<int>("error");
		var result = await start.ThenAsync(_ => ValueTask.FromResult(Failure<int>("error 2")));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task I_can_try_map_async_and_it_returns_success_for_success_without_throwing_ValueTask()
	{
		var start = Success(2);
		var result = await start.TryMapAsync(value => ValueTask.FromResult(value.ToString()));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.True);
			Assert.That(result.Value, Is.EqualTo("2"));
		});
	}

	[Test]
	public async Task I_can_try_map_async_and_it_returns_failure_for_failure_without_throwing_ValueTask()
	{
		var start = Failure<int>("error");
		var result = await start.TryMapAsync(value => ValueTask.FromResult(value.ToString()));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task I_can_try_map_async_and_it_returns_failure_for_success_with_throwing_ValueTask()
	{
		var start = Success(2);
		var result = await start.TryMapAsync(ValueTask<string> (_) => throw new CustomTestException());

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<ExceptionError>());
			var instance = result.Error as ExceptionError;
			Assert.That(instance?.Exception, Is.InstanceOf<CustomTestException>());
		});
	}

	[Test]
	public async Task I_can_try_map_async_and_it_returns_failure_for_success_with_await_throwing_ValueTask()
	{
		var start = Success(2);
		var result = await start.TryMapAsync(async ValueTask<string> (_) =>
		{
			await ValueTask.CompletedTask;
			throw new CustomTestException();
		});

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<ExceptionError>());
			var instance = result.Error as ExceptionError;
			Assert.That(instance?.Exception, Is.InstanceOf<CustomTestException>());
		});
	}

	[Test]
	public async Task I_can_try_map_async_and_it_returns_failure_for_failure_with_throwing_ValueTask()
	{
		var start = Failure<int>("error");
		var result = await start.TryMapAsync(ValueTask<string> (_) => throw new CustomTestException());

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<StringError>());
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task I_can_try_map_async_and_it_returns_failure_for_failure_with_await_throwing_ValueTask()
	{
		var start = Failure<int>("error");
		var result = await start.TryMapAsync(async ValueTask<string> (_) =>
		{
			await ValueTask.CompletedTask;
			throw new CustomTestException();
		});

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<StringError>());
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task I_can_try_transform_result_async_with_then_and_it_returns_success_for_success_and_mapping_returning_success_ValueTask()
	{
		var start = Success(2);
		var result = await start.ThenTryAsync(value => ValueTask.FromResult(Success(value.ToString())));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.True);
			Assert.That(result.Value, Is.EqualTo("2"));
		});
	}

	[Test]
	public async Task I_can_try_transform_result_async_with_then_and_it_returns_failure_for_success_and_mapping_returning_failure_ValueTask()
	{
		var start = Success(2);
		var result = await start.ThenTryAsync(_ => ValueTask.FromResult(Failure<string>("error")));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task I_can_try_transform_result_async_with_then_and_it_returns_failure_for_failure_and_mapping_returning_failure_ValueTask()
	{
		var start = Failure<int>("error");
		var result = await start.ThenTryAsync(x => ValueTask.FromResult(Success(x.ToString())));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task I_can_try_transform_result_async_with_then_and_it_returns_failure_for_success_and_mapping_throwing_ValueTask()
	{
		var start = Success(2);
		var result = await start.ThenTryAsync(ValueTask<Result<string>> (_) => throw new CustomTestException());

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<ExceptionError>());
			var instance = result.Error as ExceptionError;
			Assert.That(instance?.Exception, Is.InstanceOf<CustomTestException>());
		});
	}

	[Test]
	public async Task I_can_try_transform_result_async_with_then_and_it_returns_failure_for_success_and_mapping_await_throwing_ValueTask()
	{
		var start = Success(2);
		var result = await start.ThenTryAsync(async ValueTask<Result<string>> (_) =>
		{
			await ValueTask.CompletedTask;
			throw new CustomTestException();
		});

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<ExceptionError>());
			var instance = result.Error as ExceptionError;
			Assert.That(instance?.Exception, Is.InstanceOf<CustomTestException>());
		});
	}

	[Test]
	public async Task I_can_try_transform_result_async_with_then_and_it_returns_failure_for_failure_and_mapping_throwing_ValueTask()
	{
		var start = Failure<int>("error");
		var result = await start.ThenTryAsync(ValueTask<Result<string>> (_) => throw new CustomTestException());

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<StringError>());
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task I_can_try_transform_result_async_with_then_and_it_returns_failure_for_failure_and_mapping_await_throwing_ValueTask()
	{
		var start = Failure<int>("error");
		var result = await start.ThenTryAsync(async ValueTask<Result<string>> (_) =>
		{
			await ValueTask.CompletedTask;
			throw new CustomTestException();
		});

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<StringError>());
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}
}