// Copyright (c) The Geekeey Authors
// SPDX-License-Identifier: EUPL-1.2

using System.Diagnostics.Contracts;

namespace Geekeey.Extensions.Result;

public readonly partial struct Result<T>
{
	/// <summary>
	/// Matches over the success value or failure value of the result and returns another value. Can be conceptualized
	/// as an exhaustive <c>switch</c> expression matching all possible values of the type.
	/// </summary>
	/// <typeparam name="TResult">The type to return from the match.</typeparam>
	/// <param name="success">The function to apply to the success value of the result if the result is a success.</param>
	/// <param name="failure">The function to apply to the failure value of the result if the result is a failure.</param>
	/// <returns>The result of applying either <paramref name="success"/> or <paramref name="failure"/> on the success
	/// value or failure value of the result.</returns>
	[Pure]
	public TResult Match<TResult>(Func<T, TResult> success, Func<Error, TResult> failure)
		=> _success ? success(Value!) : failure(Error!);

	/// <summary>
	/// Matches over the success value or failure value of the result and invokes an effectful action onto the success
	/// value or failure value. Can be conceptualized as an exhaustive <c>switch</c> statement matching all possible
	/// values of the type.
	/// </summary>
	/// <param name="success">The function to call with the success value of the result if the result is a success.</param>
	/// <param name="failure">The function to call with the failure value of the result if the result is a failure.</param>
	public void Switch(Action<T> success, Action<Error> failure)
	{
		if (_success) success(Value!);
		else failure(Error!);
	}
}

public readonly partial struct Result<T>
{
	/// <summary>
	/// Asynchronously matches over the success value or failure value of the result and returns another value. Can be
	/// conceptualized as an exhaustive <c>switch</c> expression matching all possible values of the type.
	/// </summary>
	/// <typeparam name="TResult">The type to return from the match.</typeparam>
	/// <param name="success">The function to apply to the success value of the result if the result is a success.</param>
	/// <param name="failure">The function to apply to the failure value of the result if the result is a failure.</param>
	/// <returns>A task completing with the result of applying either <paramref name="success"/> or
	/// <paramref name="failure"/> on the success value or failure value of the result.</returns>
	[Pure]
	public async ValueTask<TResult> MatchAsync<TResult>(Func<T, ValueTask<TResult>> success, Func<Error, ValueTask<TResult>> failure)
		=> _success ? await success(Value!) : await failure(Error!);

	/// <summary>
	/// Asynchronously matches over the success value or failure value of the result and invokes an effectful action
	/// onto the success value or the failure value. Can be conceptualized as an exhaustive <c>switch</c> statement
	/// matching all possible values of the type.
	/// </summary>
	/// <param name="success">The function to call with the success value of the result if the result is a success.</param>
	/// <param name="failure">The function to call with the failure value of the result if the result is a failure.</param>
	public async ValueTask SwitchAsync(Func<T, ValueTask> success, Func<Error, ValueTask> failure)
	{
		if (_success) await success(Value!);
		else await failure(Error!);
	}
}

public readonly partial struct Result<T>
{
	/// <summary>
	/// Asynchronously matches over the success value or failure value of the result and returns another value. Can be
	/// conceptualized as an exhaustive <c>switch</c> expression matching all possible values of the type.
	/// </summary>
	/// <typeparam name="TResult">The type to return from the match.</typeparam>
	/// <param name="success">The function to apply to the success value of the result if the result is a success.</param>
	/// <param name="failure">The function to apply to the failure value of the result if the result is a failure.</param>
	/// <returns>A task completing with the result of applying either <paramref name="success"/> or
	/// <paramref name="failure"/> on the success value or failure value of the result.</returns>
	[Pure]
	public async Task<TResult> MatchAsync<TResult>(Func<T, Task<TResult>> success, Func<Error, Task<TResult>> failure)
		=> _success ? await success(Value!) : await failure(Error!);

	/// <summary>
	/// Asynchronously matches over the success value or failure value of the result and invokes an effectful action
	/// onto the success value or failure value. Can be conceptualized as an exhaustive <c>switch</c> statement matching
	/// all possible values of
	/// the type.
	/// </summary>
	/// <param name="success">The function to call with the success value of the result if the result is a success.</param>
	/// <param name="failure">The function to call with the failure value of the result if the result is a failure.</param>
	public async Task SwitchAsync(Func<T, Task> success, Func<Error, Task> failure)
	{
		if (_success) await success(Value!);
		else await failure(Error!);
	}
}